clc
clearvars
close all

addpath('../lib/d-stem/src'); %D-STEM software used for warning time spatial interpolation

% Change flags to 1 to enable the recomputing of the outputs of the data
% analysis
flag_intensity_map_reading=0;
flag_waveform_analysis=0;
flag_warningtime_modelling=0;
flag_warningtime_distribution_analysis=0;

%% Parameters of the M7.8 6 February 2023 earthquake
% Source: USGS
% Link: https://earthquake.usgs.gov/earthquakes/eventpage/us6000jllz/executive
% Retrival date: 31/01/2024

eq_lat = 37.226;
eq_lon = 37.014;
eq_mag = 7.8;
eq_depth = 10;
eq_origin_time = datetime('2023-02-06 01:17:34','TimeZone','UTC');

%% Data loading

%MATLAB matrices and structures
load ../input_data/population_map.mat
load ../input_data/eqn_user_coordinates.mat %to preserve privacy, a random error of +/- 20m has been added to the user locations
load ../input_data/EQN_detections/eqn_detections_fitting
load ../input_data/EQN_detections/eqn_detections_validation

%shapefiles
fault_shape = shaperead('../shapefiles/fault_shape.shp');
globe_shape = shaperead('../shapefiles/globe_shape.shp');
[globe_shape.Lon] = globe_shape.X; globe_shape = orderfields(globe_shape,[1:2,13,3:12]); globe_shape = rmfield(globe_shape,'X');
[globe_shape.Lat] = globe_shape.Y; globe_shape = orderfields(globe_shape,[1:3,13,4:12]); globe_shape = rmfield(globe_shape,'Y');

%EQN M7.8 detection data
load ../input_data/EQN_feb06_detection/triggers.mat %smartphones that detected the earthquake
load ../input_data/EQN_feb06_detection/active_smartphones.mat %monitoring smartphones at the time of the EQN earthquake detection

%Velocity model
load ../input_data/ak135/phases.mat

%% Data analysis

%EQN earthquake parameters estimate and alerting radius
beta = regress(eqn_detections_fitting.smartphone_median_acceleration-0.05,[exp(eqn_detections_fitting.catalogue_magnitude)]);

eqn_detection_time = max(triggers.timestamp);

weight = zeros(height(triggers),1);
for i=1:height(triggers)
    d = distdim(distance(triggers.lat(i),triggers.lon(i),active_smartphones.lat,active_smartphones.lon),'deg','km');
    weight(i) = 1/(max(1,sum(d<15)));
end

eqn_lat = (triggers.lat'*weight)/sum(weight);
eqn_lon = (triggers.lon'*weight)/sum(weight);
eqn_detection_delay = seconds(eqn_detection_time-eq_origin_time);
eqn_epicentre_error = distdim(distance(eq_lat,eqn_lon,eqn_lat,eqn_lon),'deg','km');
median_smartphone_acceleration = median(triggers.peak_acceleration);
eqn_magnitude_estimate = log((median_smartphone_acceleration - 0.05)/beta);
R = 6371; %Earth radius in km
depth = 10; %Assumed preliminary earthquake depth in km
radius_strong_shaking = 10^((2.31-5+1.03*eqn_magnitude_estimate)/2.15);
radius_strong_shaking = 2*R*(sqrt((radius_strong_shaking^2-depth^2)/(4*R*(R-depth))));
radius_moderate_shaking = 10^((2.31-4+1.03*eqn_magnitude_estimate)/2.15);
radius_moderate_shaking = 2*R*(sqrt((radius_moderate_shaking^2-depth^2)/(4*R*(R-depth))));
radius_mild_shaking = 10^((2.31-2+1.03*eqn_magnitude_estimate)/2.15);
radius_mild_shaking = 2*R*(sqrt((radius_mild_shaking^2-depth^2)/(4*R*(R-depth))));

disp(['EQN detection delay: ',num2str(eqn_detection_delay),' seconds'])
disp(['EQN epicentre error: ',num2str(eqn_epicentre_error),' km']);
disp(['EQN median smartphone acceleration: ',num2str(median_smartphone_acceleration),' m/s²'])
disp(['EQN estimated magnitude: ',num2str(eqn_magnitude_estimate)]);
disp(['Radius of expected strong shaking: ',num2str(radius_strong_shaking),' km']);
disp(['Radius of expected moderate shaking: ',num2str(radius_moderate_shaking),' km']);
disp(['Radius of expected mild shaking: ',num2str(radius_mild_shaking),' km']);

if flag_intensity_map_reading
    %% Intensity map loading
    % Intensity map is provided as a shapefile. The shapefile is converted
    % into a MATLAB structure aligned with the population map

    intensity_map_shape = shaperead('../input_data/macroseismic_intensity_map/M78_intensity_map.shp');

    lat_map = [];
    lon_map = [];
    intensity = [];

    for i=1:length(intensity_map_shape)
        lat_map=cat(1,lat_map,intensity_map_shape(i).YCENTER);
        lon_map=cat(1,lon_map,intensity_map_shape(i).XCENTER);
        intensity=cat(1,intensity,intensity_map_shape(i).intens);
    end

    lat_min=min(lat_map);
    lat_max=max(lat_map);
    lon_min=min(lon_map);
    lon_max=max(lon_map);

    intensity=reshape(intensity,194,257);
    intensity=flipud(intensity);

    step_lat = (lat_max-lat_min)/size(intensity,1);
    step_lon = (lon_max-lon_min)/size(intensity,2);

    lon = lon_min+step_lon/2:step_lon:lon_max-step_lon/2;
    lat = lat_max-step_lat/2:-step_lat:lat_min+step_lat/2;

    [lon,lat] = meshgrid(lon,lat);

    %intensity at the locations of EQN users
    intensity_eqn_users = griddata(lon,lat,intensity,eqn_user_coordinates.lon,eqn_user_coordinates.lat);
    %intensity interpolation to match the population grid
    intensity = griddata(lon,lat,intensity,population_map.lon,population_map.lat);

    intensity_map.intensity=intensity;
    intensity_map.lat=population_map.lat;
    intensity_map.lon=population_map.lon;

    save ../output_data/intensity_map intensity_map
    save ../output_data/intensity_eqn_users intensity_eqn_users
else
    load ../output_data/intensity_map
    load ../output_data/intensity_eqn_users
end

if flag_waveform_analysis
    % AFAD waveforms reading
    f=dir('../input_data/AFAD_waveforms/*.asc');

    counter=1;
    for i=1:3:length(f)
        S1=readlines(['../input_data/AFAD_waveforms/',f(i).name]); %east-west component
        S2=readlines(['../input_data/AFAD_waveforms/',f(i+1).name]); %north-south component
        S3=readlines(['../input_data/AFAD_waveforms/',f(i+2).name]); %up-down component

        s=S1(15);
        idx=strfind(s,": ");
        station_code=s{1}(idx+2:end);

        s=S1(17);
        idx=strfind(s,": ");
        station_lat=str2double(s{1}(idx+2:end));
        s=S1(18);
        idx=strfind(s,": ");
        station_lon=str2double(s{1}(idx+2:end));
        s=S1(25);
        idx=strfind(s,": ");
        sep=str2double(s{1}(idx+2:end));

        s=S1(27);
        idx=strfind(s,": ");
        date_fist_sample=s{1}(idx+2:end);
        date_fist_sample=datetime(date_fist_sample,'InputFormat','uuuuMMdd_HHmmss','TimeZone','UTC');

        s=S1(29);
        idx=strfind(s,": ");
        sampling_freq=str2double(s{1}(idx+2:end));

        w1=str2double(S1(65:end));
        w2=str2double(S2(65:end));
        w3=str2double(S3(65:end));

        pga=max(abs(w1),abs(w2)); %max of NS and EW components

        idx=find(pga>117.72,1); %12%g in cm/s^2
        if not(isempty(idx))
            delta_s=idx*sampling_freq;

            date_pick=date_fist_sample+seconds(delta_s);
            delta_pick=seconds(date_pick-eq_origin_time);
            if delta_pick<0
                delta_pick=NaN;
            end
        else
            delta_pick=NaN;
        end

        waveform_data{counter}.station_code=station_code;
        waveform_data{counter}.lat=station_lat;
        waveform_data{counter}.lon=station_lon;
        waveform_data{counter}.separation=sep;
        waveform_data{counter}.delta_pick=delta_pick;
        waveform_data{counter}.acceleration=w1;

        delta_t=seconds(sampling_freq);
        date_last_sample=date_fist_sample+delta_t*(length(w1)-1);
        t=date_fist_sample:delta_t:date_last_sample;
        waveform_data{counter}.t=t;

        counter=counter+1;
    end

    waveform_data=cell2mat(waveform_data);
    waveform_data=struct2table(waveform_data);
    
    save ../output_data/waveform_data waveform_data
else
    load ../output_data/waveform_data
end

if flag_warningtime_modelling
    % warning time modelling using the D-STEM software
    idx=find(not(isnan(waveform_data.delta_pick)));
    [b,~,r]=regress(waveform_data.delta_pick(idx),[ones(length(idx),1) waveform_data.separation(idx)]);

    shp = alphaShape(waveform_data.lon(idx),waveform_data.lat(idx),'HoleThreshold',0);
    a = criticalAlpha(shp,'all-points');
    shp = alphaShape(waveform_data.lon(idx),waveform_data.lat(idx),a*2,'HoleThreshold',0);
    [bf,boundary_pga] = boundaryFacets(shp);

    %% D-STEM
    y=waveform_data.delta_pick(idx);
    ground.Y{1} = y;
    ground.Y_name{1} = 'pick time';
    n = size(ground.Y{1}, 1);
    T = size(ground.Y{1}, 2);

    %load covariates for the NO2 monitoring stations
    ground.X_beta{1} = [ones(length(idx),1) waveform_data.separation(idx)];
    ground.X_beta_name{1} = {'constant','separation'};

    ground.X_p{1} = ones(length(idx),1);
    ground.X_p_name{1} = {'constant'};

    obj_stem_varset_p = stem_varset(ground.Y, ground.Y_name, [], [], ...
        ground.X_beta, ground.X_beta_name, ...
        [], [], ...
        ground.X_p,ground.X_p_name);

    %Coordinates
    obj_stem_gridlist_p = stem_gridlist();
    ground.coordinates{1} = [waveform_data.lat(idx), waveform_data.lon(idx)];
    obj_stem_grid = stem_grid(ground.coordinates{1}, 'deg', 'sparse', 'point');
    obj_stem_gridlist_p.add(obj_stem_grid);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %      Model building     %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    obj_stem_datestamp = stem_datestamp('01-01-2009 00:00', '01-01-2009 00:00', T);

    %stem_data object creation
    obj_stem_modeltype = stem_modeltype('DCM');
    obj_stem_data = stem_data(obj_stem_varset_p, obj_stem_gridlist_p, ...
        [], [], obj_stem_datestamp, [], obj_stem_modeltype);
    %stem_par object creation
    obj_stem_par_constraints=stem_par_constraints();
    obj_stem_par = stem_par(obj_stem_data, 'exponential',obj_stem_par_constraints);
    %stem_model object creation
    obj_stem_model = stem_model(obj_stem_data, obj_stem_par);

    %Starting values
    obj_stem_par.beta = b;
    obj_stem_par.theta_p = 15; %km
    obj_stem_par.v_p = var(r)*0.7;
    obj_stem_par.sigma_eps = var(r)*0.3;

    obj_stem_model.set_initial_values(obj_stem_par);

    %Model estimation
    obj_stem_EM_options = stem_EM_options();
    obj_stem_EM_options.exit_tol_par=0.005;
    obj_stem_EM_options.max_iterations=200;
    obj_stem_model.EM_estimate(obj_stem_EM_options);

    %% Kriging

    mask=inpolygon(population_map.lon(:),population_map.lat(:),boundary_pga(:,1),boundary_pga(:,2));
    mask=reshape(mask,size(population_map.lat));
    mask=double(mask);
    mask(mask==0)=NaN;

    krig_separation=distdim(distance(population_map.lat(:),population_map.lon(:),eq_lat,eq_lon),'deg','km');

    obj_stem_krig_grid = stem_grid([population_map.lat(:),population_map.lon(:)], 'deg', 'regular','pixel',size(population_map.lat),'square',0.75,0.75);

    obj_stem_krig_data = stem_krig_data(obj_stem_krig_grid,[ones(length(krig_separation),1) krig_separation],{'constant','separation'},mask);
    obj_stem_krig = stem_krig(obj_stem_model,obj_stem_krig_data);

    obj_stem_krig_options = stem_krig_options();
    obj_stem_krig_options.block_size = 500;

    obj_stem_krig_result = obj_stem_krig.kriging(obj_stem_krig_options);

    pick_map=obj_stem_krig_result{1};
    mask_kriging=mask;

    warning_time_map=pick_map.y_hat-eqn_detection_delay;
    warning_time_eqn_users = griddata(population_map.lon,population_map.lat,warning_time_map,eqn_user_coordinates.lon,eqn_user_coordinates.lat);

    save ../output_data/obj_stem_model obj_stem_model
    save ../output_data/mask_kriging mask_kriging
    save ../output_data/boundary_pga boundary_pga
    save ../output_data/warning_time_map warning_time_map
    save ../output_data/warning_time_eqn_users warning_time_eqn_users
else
    load ../output_data/obj_stem_model 
    load ../output_data/mask_kriging
    load ../output_data/boundary_pga
    load ../output_data/warning_time_map
    load ../output_data/warning_time_eqn_users
end

if flag_warningtime_distribution_analysis
    intensity_map.intensity=round(intensity_map.intensity);

    mask=mask_kriging;
    mask(isnan(mask))=false;
    mask(mask==1)=true;

    separation = distdim(distance(eqn_lat,eqn_lon,population_map.lat,population_map.lon),'deg','km');
    mask_strong = mask & separation<=radius_strong_shaking;
    mask_moderate = mask & separation>radius_strong_shaking & separation<radius_moderate_shaking;

    L=intensity_map.intensity>=9 & mask_strong;
    wt=warning_time_map(L);
    pop_wt=population_map.count(L);
    warning_time_distribution_IX_and_higher=[];
    for h=1:length(wt)
        if not(isnan(pop_wt(h)))
            warning_time_distribution_IX_and_higher=cat(1,warning_time_distribution_IX_and_higher,repmat(wt(h),round(pop_wt(h)),1));
        end
    end

    L=intensity_map.intensity==8 & mask_strong;
    wt=warning_time_map(L);
    pop_wt=population_map.count(L);
    warning_time_distribution_VIII=[];
    for h=1:length(wt)
        if not(isnan(pop_wt(h)))
            warning_time_distribution_VIII=cat(1,warning_time_distribution_VIII,repmat(wt(h),round(pop_wt(h)),1));
        end
    end

    L=intensity_map.intensity>=8 & mask_moderate;
    wt=warning_time_map(L);
    pop_wt=population_map.count(L);
    warning_time_distribution_VIII_radius_moderate=[];
    for h=1:length(wt)
        if not(isnan(pop_wt(h)))
            warning_time_distribution_VIII_radius_moderate=cat(1,warning_time_distribution_VIII_radius_moderate,repmat(wt(h),round(pop_wt(h)),1));
        end
    end

    save ../output_data/warning_time_distribution_VIII warning_time_distribution_VIII;
    save ../output_data/warning_time_distribution_IX_and_higher warning_time_distribution_IX_and_higher;
else
    load ../output_data/warning_time_distribution_VIII;
    load ../output_data/warning_time_distribution_IX_and_higher;
end

%% Article figures
[cmap_delay,cmap_warning,cmap_intensity,cmap_population,cmap_waveforms,cmap_acceleration] = load_colormaps();

stations_code={'4620','4615','4616','2712','3144','3142','3129','2718','2703'};
stations_city_name={'Kahramanmaraş','İsmetpaşa','Türkoğlu','Nurdağı','Hassa','Kırıkhan','Antakya','İslahiye','Gaziantep'};
station_marker={'o','+','*','x','square','diamond','v','>','<'};

idx_selected=[];
idx_unselected=[];
for i=1:height(waveform_data)
    code{i}=waveform_data.station_code{i};
    if sum(strcmp(code{i},stations_code))
        idx_selected=cat(1,idx_selected,i);
    else
        idx_unselected=cat(1,idx_unselected,i);
    end
end
waveform_data_selected=waveform_data(idx_selected,:);
waveform_data_unselected=waveform_data(idx_unselected,:);
for i=1:height(waveform_data_selected)
    sep(i,1)=waveform_data_selected.separation(i);
    idx=strcmp(waveform_data_selected.station_code{i},stations_code);
    waveform_data_selected.stations_city_name{i}=[stations_city_name{idx},' station'];
end
[v,idx]=sort(sep);
waveform_data_selected=waveform_data_selected(idx,:);

%% Figure 1
subplot(1,3,[1 2])
[lat_cP,lon_cP] = scircle1(eq_lat,eq_lon,distdim(67.71,'km','deg'));
a0=geoplot(lat_cP,lon_cP,'LineWidth',3.3,'Color',[1 0.7840 0]);
hold on
[lat_cS,lon_cS] = scircle1(eq_lat,eq_lon,distdim(39.96,'km','deg'));
a1=geoplot(lat_cS,lon_cS,'LineWidth',3.3,'Color',[0.4940 0.1840 0.5560]);
a1b=geoplot(eqn_user_coordinates.lat,eqn_user_coordinates.lon,'o','MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',5,'LineWidth',1);
a2=geoplot(active_smartphones.lat,active_smartphones.lon,'o','LineWidth',1,'MarkerEdgeColor','k','MarkerSize',12,'LineWidth',1.5);%[0.3 0.3 0.3]
trigger_delay=seconds(triggers.timestamp-eq_origin_time);
counter=1;
for t=7.5:0.5:12.5
    L=trigger_delay>=t-0.5&trigger_delay<=t;
    if counter==1
        a3=geoplot(triggers.lat(L),triggers.lon(L),'o','LineWidth',1,'MarkerEdgeColor',[0 0 0],'MarkerFaceColor',cmap_delay(counter,:),'MarkerSize',15,'LineWidth',1.8);
    else
        geoplot(triggers.lat(L),triggers.lon(L),'o','LineWidth',1,'MarkerEdgeColor',[0 0 0],'MarkerFaceColor',cmap_delay(counter,:),'MarkerSize',15,'LineWidth',1.8);
    end
    counter=counter+1;
end
for j=1:length(fault_shape)
    if j==1
        a4=geoplot(fault_shape(j).Y,fault_shape(j).X,'Color','b','LineWidth',2.5);
    else
        geoplot(fault_shape(j).Y,fault_shape(j).X,'Color','b','LineWidth',2.5);
    end
end
a5=geoplot(eq_lat,eq_lon,'p','MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',24,'LineWidth',1.5);
a6=geoplot(eqn_lat,eqn_lon,'p','MarkerEdgeColor','k','MarkerFaceColor',[0.3010 0.7450 0.9330],'MarkerSize',24,'LineWidth',1.5);
c=colorbar;
c.Label.String='Trigger delay (s)';
c.LineWidth=2;
c.FontWeight='normal';
c.FontSize=20;
clim([7,12.5])
colormap(gca,cmap_delay);
set(gca,'FontSize',22);
set(gca,'LineWidth',2.4);
set(gca,'FontName','Arial');
legend([a0,a1,a1b,a2,a3,a4,a5,a6],{'P-wave at EQN detection','S-wave at EQN detection','EQN user','Active smartphone','Triggering smartphone','Fault rupture','M7.8 epicentre','EQN preliminary epicentre'},'Location','northwest','FontSize',18,'FontWeight','normal');
title('A','FontWeight','bold','FontName','Arial');

subplot(1,3,3)

idx = Sphase.depth==eq_depth;
Sphase_small=Sphase(idx,:);
idx = Pphase.depth==eq_depth;
Pphase_small=Pphase(idx,:);
distance_triggers = distdim(distance(eq_lat,eq_lon,triggers.lat,triggers.lon),'deg','km');
d_vec=0:10:95;
t_S=[];
t_P=[];
for i=1:length(d_vec)
    idx2=find(d_vec(i)<Sphase_small.distance,1);
    idx1=idx2-1;
    d_1=Sphase_small.distance(idx1);
    d_2=Sphase_small.distance(idx2);
    r=(d_vec(i)-d_1)/(d_2-d_1);
    t_1=Sphase_small.time(idx1);
    t_2=Sphase_small.time(idx2);
    t_S(i)=t_1+r*(t_2-t_1);

    idx2=find(d_vec(i)<Pphase_small.distance,1);
    idx1=idx2-1;
    d_1=Pphase_small.distance(idx1);
    d_2=Pphase_small.distance(idx2);
    r=(d_vec(i)-d_1)/(d_2-d_1);
    t_1=Sphase_small.time(idx1);
    t_2=Sphase_small.time(idx2);
    t_P(i)=t_1+r*(t_2-t_1);
end

plot(d_vec,t_P,'LineWidth',3.3,'Color',[1 0.7840 0])
hold on
plot(d_vec,t_S,'LineWidth',3.3,'Color',[0.4940 0.1840 0.5560])
plot([0,100],[seconds(eqn_detection_time-eq_origin_time) seconds(eqn_detection_time-eq_origin_time)],'m--','LineWidth',2.5);
scatter(distance_triggers,seconds(triggers.timestamp-eq_origin_time),135,triggers.peak_acceleration,'filled','MarkerEdgeColor','k','LineWidth',1.5);
c=colorbar;
c.Ticks=0.5:0.3:3.5;
c.FontWeight='normal';
c.FontSize=20;
c.LineWidth=2;
clim([0.5,3.5])
c.Label.String='Smartphone peak resultant acceleration (m·s^{-2})';
xlabel('Epicentral distance (km)');
ylabel('Delay from origin time (s)');
legend({'Expected P-wave arrival','Expected S-wave arrival','EQN detection','Smartphone trigger'},'Location','northwest','FontWeight','normal','FontSize',18);
xlim([20,45]);
xticks(20:5:45);
set(gca,'FontSize',22);
%set(gca, 'FontWeight', 'bold');
set(gca,'LineWidth',2.4);
set(gca,'FontName','Arial')
set(gca,'GridLineStyle','--')
grid on
box on
colormap(gca,cmap_acceleration);
title('B','FontWeight','bold','FontName','Arial');

%% Figure 2
figure

boundary_pga=cat(1,boundary_pga,boundary_pga(1,:));
[lat_c,lon_c] = scircle1(eqn_lat,eqn_lon,distdim(150,'km','deg'));

warning_time_map(warning_time_map<0)=NaN;
L = find(not(isnan(intensity_map.intensity))); 
[idx_r,idx_c]=ind2sub(size(intensity_map.intensity),L);
R = georefcells([min(population_map.lat(L)),max(population_map.lat(L))],[min(population_map.lon(L)),max(population_map.lon(L))],size(warning_time_map(min(idx_r):max(idx_r),min(idx_c):max(idx_c))),'ColumnsStartFrom','north');
wm=worldmap(intensity_map.intensity(min(idx_r):max(idx_r),min(idx_c):max(idx_c)),R);
wm.LineWidth=2.4;

objs = findobj(wm, 'type', 'text');
for i=1:length(objs)
    objs(i).FontName='Arial';
    objs(i).FontSize=20;
end
objs = findobj(wm, 'type', 'line');
for i=1:length(objs)
    objs(i).LineWidth=1;
    objs(i).LineStyle='--';
end
objs = findobj(wm, 'type', 'patch');
for i=1:length(objs)
    objs(i).LineWidth=2.4;
end

set(gca,'FontName','Arial','FontSize',18)
h=geoshow(warning_time_map(min(idx_r):max(idx_r),min(idx_c):max(idx_c)),R,'DisplayType','texturemap');
set(h,'FaceColor','flat');
colormap(gca,cmap_warning);
p1=geoshow(boundary_pga(:,2),boundary_pga(:,1),'DisplayType','line','LineWidth',3,'Color','r');
p2=geoshow(waveform_data_unselected.lat,waveform_data_unselected.lon,'DisplayType','multipoint','MarkerEdgeColor','b','Marker','^','LineWidth',2.5,'MarkerSize',13);
for j=1:height(waveform_data_selected)
    p3{j}=geoshow(waveform_data_selected.lat(j),waveform_data_selected.lon(j),'DisplayType','point','MarkerEdgeColor',[0.95 0 0],'Marker',station_marker{j},'LineWidth',1.75,'MarkerSize',13);
end

for j=1:length(fault_shape)
    if j==1
        p4 = geoshow(fault_shape(j).Y,fault_shape(j).X,'DisplayType','line','Color','k','LineWidth',2.5);
    else
        geoshow(fault_shape(j).Y,fault_shape(j).X,'DisplayType','line','Color','k','LineWidth',2.5);
    end
end
p5=geoshow(eq_lat,eq_lon,'DisplayType','point','Marker','p','MarkerSize',28,'MarkerEdgeColor','k','MarkerFaceColor','r','LineWidth',2.5);
p6=geoshow(eqn_lat,eqn_lon,'DisplayType','point','Marker','p','MarkerSize',28,'MarkerEdgeColor','k','MarkerFaceColor',[0.3010 0.7450 0.9330],'LineWidth',2.5);
p7=geoshow(lat_c,lon_c,'DisplayType','line','LineStyle','--','Color','k','LineWidth',2);
p8=geoshow(globe_shape,'LineWidth',3,'Color','k');
clim([0 70]);
box on
c=colorbar;
c.Label.String='Warning time w.r.t. PGA>12%g (s)';
c.LineWidth=1.8;
c.FontWeight='normal';
c.FontSize=18;
set(gca,'LineWidth',2.4);
set(gca,'FontName','Arial');
legend([p8,p1,p4,p7,p6,p5,p3{1},p3{2},p3{3},p3{4},p3{5},p3{6},p3{7},p3{8},p3{9},p2],['Turkish-Syrian border','PGA>12%g area','Fault rupture','Strong shaking alert area','EQN preliminary epicentre','M7.8 epicentre',waveform_data_selected.stations_city_name(:)',{'Other AFAD stations'}],'FontWeight','normal','FontSize',18);

%% Figure 3
figure
subplot(1,2,1)
wm=worldmap(intensity_map.intensity(min(idx_r):max(idx_r),min(idx_c):max(idx_c)),R);
objs = findobj(wm, 'type', 'text');
for i=1:length(objs)
    objs(i).FontName='Arial';
    objs(i).FontSize=22;
end
objs = findobj(wm, 'type', 'line');
for i=1:length(objs)
    objs(i).LineWidth=1;
    objs(i).LineStyle='--';
end
objs = findobj(wm, 'type', 'patch');
for i=1:length(objs)
    objs(i).LineWidth=2.8;
end

h=geoshow(intensity_map.intensity(min(idx_r):max(idx_r),min(idx_c):max(idx_c)),R,'DisplayType','texturemap');
set(h,'FaceColor','flat');
hold on
geoshow(boundary_pga(:,2),boundary_pga(:,1),'DisplayType','line','LineWidth',2,'Color','r');
for j=1:length(fault_shape)
    geoshow(fault_shape(j).Y,fault_shape(j).X,'DisplayType','line','Color','k','LineWidth',1.5);
end
geoshow(eq_lat,eq_lon,'DisplayType','point','Marker','p','MarkerSize',24,'MarkerEdgeColor','k','MarkerFaceColor','r','LineWidth',1.5);
geoshow(eqn_lat,eqn_lon,'DisplayType','point','Marker','p','MarkerSize',24,'MarkerEdgeColor','k','MarkerFaceColor',[0.3010 0.7450 0.9330],'LineWidth',1.5);
geoshow(lat_c,lon_c,'DisplayType','line','LineStyle','--','Color','k','LineWidth',2);
geoshow(globe_shape,'LineWidth',2.5,'Color','k');
clim([4.5,11.5]);
box on
c=colorbar('Ticks',5:11);
c.TickLabels={'V','VI','VII','VIII','IX','X','XI'};
c.Label.String='Macroseismic intensity';
c.LineWidth=2;
c.FontWeight='normal';
c.FontSize=22;
colormap(gca,cmap_intensity)
xlabel('Longitude')
ylabel('Latitude')
title('A');
set(gca,'FontSize',22);

subplot(1,2,2)
wm=worldmap(population_map.count(min(idx_r):max(idx_r),min(idx_c):max(idx_c)),R);
objs = findobj(wm, 'type', 'text');
for i=1:length(objs)
    objs(i).FontName='Arial';
    objs(i).FontSize=22;
end
objs = findobj(wm, 'type', 'line');
for i=1:length(objs)
    objs(i).LineWidth=1;
    objs(i).LineStyle='--';
end
objs = findobj(wm, 'type', 'patch');
for i=1:length(objs)
    objs(i).LineWidth=2.8;
end

h=geoshow(log10(population_map.count(min(idx_r):max(idx_r),min(idx_c):max(idx_c))),R,'DisplayType','texturemap');
set(h,'FaceColor','flat');
geoshow(boundary_pga(:,2),boundary_pga(:,1),'DisplayType','line','LineWidth',2,'Color','r');
for j=1:length(fault_shape)
    geoshow(fault_shape(j).Y,fault_shape(j).X,'DisplayType','line','Color','k','LineWidth',1.5);
end
geoshow(eq_lat,eq_lon,'DisplayType','point','Marker','p','MarkerSize',24,'MarkerEdgeColor','k','MarkerFaceColor','r','LineWidth',1.5);
geoshow(eqn_lat,eqn_lon,'DisplayType','point','Marker','p','MarkerSize',24,'MarkerEdgeColor','k','MarkerFaceColor',[0.3010 0.7450 0.9330],'LineWidth',1.5);
geoshow(lat_c,lon_c,'DisplayType','line','LineStyle','--','Color','k','LineWidth',2);
geoshow(globe_shape,'LineWidth',2.5,'Color','k');
clim([0,5]);
box on
c=colorbar('Ticks',0:5);
c.TickLabels={'10^0','10^1','10^2','10^3','10^4','10^5'};
c.Label.String='Population count';
c.LineWidth=2;
c.FontWeight='normal';
c.FontSize=22;
colormap(gca,cmap_population)
title('B');
set(gca,'FontSize',22);

%% Figure 4
figure;

hold on
for i=1:height(waveform_data_selected)
    t=waveform_data_selected.t{i};
    t=[datetime('2023-02-06 01:17:00','TimeZone','UTC'),t,datetime('2023-02-06 01:20:00','TimeZone','UTC')];
    acc=waveform_data_selected.acceleration{i};
    acc(1)=0;
    acc(end)=0;
    acc(isnan(acc))=0;
    acc=[0;acc;0];
    idx = find(abs(waveform_data_selected.acceleration{i})>117.72,1); %12%g in cm/s^2
    if not(isempty(idx))
        plot(t,acc/1400+i,'-','Color',cmap_waveforms(i,:),'LineWidth',1.0);
        if i==1
            p3=plot(t(idx),i,'ok','MarkerSize',8,'LineWidth',2);
        else
            plot(t(idx),i,'ok','MarkerSize',8,'LineWidth',2);
        end
    end
    sep_label{i}=num2str(round(waveform_data_selected.separation(i)));
    station_code_label{i}=waveform_data_selected.station_code{i};
    idx=find(strcmp(station_code_label{i},stations_code));
    station_code_label{i}=[station_code_label{i},'\newline',stations_city_name{idx}];
end
p1=plot([eq_origin_time eq_origin_time],[0 10],'Color',[0.6350 0.0780 0.1840],'LineWidth',2.5);
p2=plot([eqn_detection_time eqn_detection_time],[0 10],'Color',[0 0.4470 0.7410],'LineWidth',2.5);
set(gca,'Ydir','reverse')
yticks(1:9);
ylim([0,10]);
xticks([eq_origin_time,eqn_detection_time,eqn_detection_time+seconds(5),eqn_detection_time+seconds(10),...
    eqn_detection_time+seconds(15),eqn_detection_time+seconds(20),eqn_detection_time+seconds(25),...
    eqn_detection_time+seconds(30),eqn_detection_time+seconds(35),eqn_detection_time+seconds(40),...
    eqn_detection_time+seconds(45),eqn_detection_time+seconds(50),eqn_detection_time+seconds(55),...
    eqn_detection_time+seconds(60),eqn_detection_time+seconds(65),eqn_detection_time+seconds(70),...
    eqn_detection_time+seconds(80),eqn_detection_time+seconds(90),eqn_detection_time+seconds(100),...
    eqn_detection_time+seconds(110),eqn_detection_time+seconds(120),eqn_detection_time+seconds(130)])

xticklabels({datestr(eq_origin_time,'HH:MM:SS'),datestr(eqn_detection_time,'HH:MM:SS'),'+5s','+10s','+15s','+20s','+25s','+30s','+35s','+40s','+45s','+50s','+55s','+60s','+65s','+70s','80s','+90s','+100s','+110s','+120s','+130s'});
yticklabels(sep_label);
ylabel('Station separation from epicentre (km)');

xlim([datetime('2023-02-06 01:17:30','TimeZone','UTC'),datetime('2023-02-06 01:19:46','TimeZone','UTC')])
box on
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');
set(gca,'FontSize',22);
set(gca,'LineWidth',2.4);
set(gca,'FontName','Arial')

yyaxis right
set(gca,'Ydir','reverse')
plot([0 0],[0 0]);
ylim([0,10]);
yticks(1:9);
yticklabels(station_code_label);
ylabel('AFAD station code and city');
set(gca,'YColor',[0 0 0])

set(gca,'FontSize',22);
set(gca,'LineWidth',2.4);
set(gca,'FontName','Arial')

legend([p1,p2,p3],{'Origin time','EQN detection','12%g PGA exceedance time'})
xlabel('Time UTC');

%% Figure 5
figure
subplot(2,2,1)
patch([-10 0 0 -10],[0 0 15^6 15^6],[0.6350 0.0780 0.1840],'FaceAlpha',0.2,'EdgeAlpha',0);
patch([0 60 60 0],[0 0 15^6 15^6],[0.4660 0.6740 0.1880],'FaceAlpha',0.2,'EdgeAlpha',0);
hold on
separation_eqn_users =  distdim(distance(eqn_lat,eqn_lon,eqn_user_coordinates.lat,eqn_user_coordinates.lon),'deg','km');
L=separation_eqn_users<=radius_strong_shaking & round(intensity_eqn_users)>8;
histogram(round(warning_time_eqn_users(L)),'BinEdges',-15+0.00001:5:70,'FaceColor',[0.4940 0.1840 0.5560],'LineWidth',1.5);
title('A');
xlabel('Warning time (s)');
ylabel('EQN users count');
xticks(-10:5:60);
xlim([-10,60]);
ylim([0 100]);
box on
set(gca,'FontSize',16)
set(gca,'LineWidth',2);
set(gca,'FontName','Arial')
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');

grid on

subplot(2,2,2)
patch([-10 0 0 -10],[0 0 15^6 15^6],[0.6350 0.0780 0.1840],'FaceAlpha',0.2,'EdgeAlpha',0);
patch([0 60 60 0],[0 0 15^6 15^6],[0.4660 0.6740 0.1880],'FaceAlpha',0.2,'EdgeAlpha',0);
hold on
L=separation_eqn_users<=141 & round(intensity_eqn_users)==8;
histogram(round(warning_time_eqn_users(L)),'BinEdges',-15+0.00001:5:70,'FaceColor',[0.4940 0.1840 0.5560],'LineWidth',1.5);
title('B');
xlabel('Warning time (s)');
ylabel('EQN users count');
xticks(-10:5:60);
xlim([-10,60]);
ylim([0 50]);
box on
set(gca,'FontSize',16)
set(gca,'LineWidth',2);
set(gca,'FontName','Arial')
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');
grid on

subplot(2,2,3)
yyaxis left
patch([-10 0 0 -10],[0 0 15^6 15^6],[0.6350 0.0780 0.1840],'FaceAlpha',0.2,'EdgeAlpha',0);
patch([0 60 60 0],[0 0 15^6 15^6],[0.4660 0.6740 0.1880],'FaceAlpha',0.2,'EdgeAlpha',0);
hold on
hh=histogram(warning_time_distribution_IX_and_higher,'BinEdges',-10:5:60,'LineWidth',1.5);
xticks(-10:5:60);
if (max(hh.Values)<10000)
    ylim([0,ceil(max(hh.Values)/1000)*1000]);
elseif (max(hh.Values)<100000)
    ylim([0,ceil(max(hh.Values)/10000)*10000]);
else
    ylim([0,ceil(max(hh.Values)/100000)*100000]);
end

xlim([-10,60]);
ylim([0,2000000]);
yticks(0:500000:2000000);
box on
xlabel('Warning time (s)');
ylabel('Population count');
title('C');
set(gca,'FontSize',16)
set(gca,'LineWidth',2);
set(gca,'FontName','Arial')
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');
c=[0 hh.Values];
c=cumsum(c);
c=c(end)-c;
yyaxis right
plot(-10:5:60,c,'.-r','LineWidth',2.4);
plot(30,c(9),'or','MarkerSize',10,'LineWidth',2)
ylim([0,4000000]);
grid on
set(gca,'YColor',[1 0 0])
ylabel('Cumulative population count');

subplot(2,2,4)
yyaxis left
patch([-10 0 0 -10],[0 0 15^6 15^6],[0.6350 0.0780 0.1840],'FaceAlpha',0.2,'EdgeAlpha',0);
patch([0 60 60 0],[0 0 15^6 15^6],[0.4660 0.6740 0.1880],'FaceAlpha',0.2,'EdgeAlpha',0);
hold on
hh=histogram(warning_time_distribution_VIII,'BinEdges',-10:5:60,'LineWidth',1.5);
xticks(-10:5:60);
if (max(hh.Values)<10000)
    ylim([0,ceil(max(hh.Values)/1000)*1000]);
elseif (max(hh.Values)<100000)
    ylim([0,ceil(max(hh.Values)/10000)*10000]);
else
    ylim([0,ceil(max(hh.Values)/100000)*100000]);
end
xlim([-10,60]);
ylim([0,2000000]);
yticks(0:500000:2000000);
box on
grid on
xlabel('Warning time (s)');
ylabel('Population count');
title('D');
set(gca,'FontSize',16)
set(gca,'LineWidth',2);
set(gca,'FontName','Arial')
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');
c=[0 hh.Values];
c=cumsum(c);
c=c(end)-c;
yyaxis right
plot(-10:5:60,c,'.-r','LineWidth',2.4);
plot(30,c(9),'or','MarkerSize',10,'LineWidth',2)
ylim([0,4000000]);
set(gca,'YColor',[1 0 0])
ylabel('Cumulative population count');

%% Figure 6
m_hat_val = log((eqn_detections_validation.smartphone_median_acceleration - 0.05)/beta);

figure
plot(eqn_detections_fitting.catalogue_magnitude,eqn_detections_fitting.smartphone_median_acceleration,'o','MarkerSize',14,'LineWidth',1.5);
hold on
plot(eqn_detections_validation.catalogue_magnitude,eqn_detections_validation.smartphone_median_acceleration ,'^','MarkerSize',14,'LineWidth',1.5);
text(eqn_detections_validation.catalogue_magnitude(51),eqn_detections_validation.smartphone_median_acceleration(51)+0.25,'M7.8','VerticalAlignment','middle','HorizontalAlignment','center','FontSize',18,'FontWeight','bold')
text(eqn_detections_validation.catalogue_magnitude(82),eqn_detections_validation.smartphone_median_acceleration(82)+0.25,'M6.3','VerticalAlignment','middle','HorizontalAlignment','center','FontSize',18,'FontWeight','bold')

mm = 2:0.01:8;
ga = exp(mm)*beta + 0.05;
plot(mm,ga,'LineWidth',2.5)

xlim([2,8]);
xticks(2:0.5:8)
grid on
xlabel('Catalogue magnitude')
ylabel('EQN median smartphone acceleration (m·s^{-2})')
set(gca,'FontSize',20);
set(gca,'LineWidth',2.4);
set(gca,'xgrid','on','GridLineWidth',1.5,'GridLineStyle','--');
legend({['Training dataset, n = ',num2str(height(eqn_detections_fitting)),', [Jan 2018–Dec 2022]'],['Validation dataset, n = ',num2str(height(eqn_detections_validation)),', [Jan 2023–May 2023]'],'Ground acceleration vs magnitude relationship'},'Location','northwest')